# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function:
import cv2
import numpy as np
from pathlib import Path
from zhousflib.decorator import interceptor_util


def read(img_path: Path):
    """
    读图片-兼容图片路径包含中文
    :param img_path:
    :return: np.ndarray
    """
    img = cv2.imdecode(np.fromfile(str(img_path), dtype=np.uint8), cv2.IMREAD_COLOR)
    return img


def write(image: np.ndarray, img_write_path: Path):
    """
    写图片-兼容图片路径包含中文
    :param image:
    :param img_write_path:
    :return:
    """
    cv2.imencode(img_write_path.suffix, image[:, :, ::-1])[1].tofile(str(img_write_path))


def max_connectivity_domain(mask_arr: np.array, connectivity=4) -> np.array:
    """
    返回掩码中最大的连通域
    :param mask_arr: 二维数组，掩码中0表示背景，1表示目标
    :param connectivity: 4|8 4邻接还是8邻接
    :return:
    """
    # 掩码标识转换
    arr_mask = np.where(mask_arr == 1, 255, 0)
    # 掩码类型转换
    arr_mask = arr_mask.astype(dtype=np.uint8)
    """
    connectivity：可选值为4或8，也就是使用4连通还是8连通
    num：所有连通域的数目
    labels：图像上每一像素的标记，用数字1、2、3…表示（不同的数字表示不同的连通域）
    stats：每一个标记的统计信息，是一个5列的矩阵，每一行对应每个连通区域的外接矩形的x、y、width、height和面积，示例：0 0 720 720 291805
    centroids：连通域的中心点
    """
    nums, labels, stats, centroids = cv2.connectedComponentsWithStats(arr_mask, connectivity=connectivity)
    background = 0
    for row in range(stats.shape[0]):
        if stats[row, :][0] == 0 and stats[row, :][1] == 0:
            background = row
    # 删除背景后的连通域列表
    stats_no_bg = np.delete(stats, background, axis=0)
    # 获取连通域最大的索引
    max_idx = stats_no_bg[:, 4].argmax()
    max_region = np.where(labels == max_idx + 1, 1, 0)
    # 保存
    # cv2.imwrite(r'vis.jpg', max_region * 255)
    return max_region


def rotate(image, angle, show=False):
    """
    顺时针旋转图片
    :param image: 图片
    :param angle: 旋转角度
    :param show: 是否显示
    :return: np.array
    """
    if isinstance(image, str):
        image = cv2.imread(image)
    if isinstance(image, np.ndarray):
        # grab the dimensions of the image and then determine the
        # center
        (h, w) = image.shape[:2]
        (cX, cY) = (w // 2, h // 2)

        # grab the rotation matrix (applying the negative of the
        # angle to rotate clockwise), then grab the sine and cosine
        # (i.e., the rotation components of the matrix)
        M = cv2.getRotationMatrix2D((cX, cY), -angle, 1.0)
        cos = np.abs(M[0, 0])
        sin = np.abs(M[0, 1])

        # compute the new bounding dimensions of the image
        nW = int((h * sin) + (w * cos))
        nH = int((h * cos) + (w * sin))

        # adjust the rotation matrix to take into account translation
        M[0, 2] += (nW / 2) - cX
        M[1, 2] += (nH / 2) - cY

        # perform the actual rotation and return the image
        rotated = cv2.warpAffine(image, M, (nW, nH))
        if show:
            cv2.imshow("{0} rotated".format(angle), rotated)
            cv2.waitKey(0)
        return rotated
    return None


def _read(chain):
    image, show = chain
    if isinstance(image, str):
        image = cv2.imread(image)
    if image is None:
        return True, image
    return False, (image, show)


def _show(chain):
    title, image, show = chain
    if show:
        cv2.imshow(title, image)
        cv2.waitKey(0)
        cv2.destroyAllWindows()
    return image


@interceptor_util.intercept(before=_read, after=_show)
def img_binary(chain):
    """ 二值化 """
    image, show = chain
    if len(image.shape) == 3:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    block_size = 11  # 分割计算的区域大小，取奇数
    binary_ = cv2.adaptiveThreshold(image, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY_INV, block_size, 15)
    return "binary", binary_, show


@interceptor_util.intercept(before=_read, after=_show)
def img_bilateral_filter(chain):
    """ 双边滤波 """
    image, show = chain
    bilateral_filter_ = cv2.bilateralFilter(image, 7, 75, 75)
    return "bilateral_filter", bilateral_filter_, show


@interceptor_util.intercept(before=_read, after=_show)
def img_erosion(chain):
    """ 腐蚀 """
    image, show = chain
    kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (20, 20))
    erosion_ = cv2.erode(image, kernel)
    return "erosion", erosion_, show


@interceptor_util.intercept(before=_read, after=_show)
def img_dilate(chain):
    """ 膨胀 """
    image, show = chain
    kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (20, 20))
    dilate_ = cv2.dilate(image, kernel)
    return "dilate", dilate_, show


@interceptor_util.intercept(before=_read, after=_show)
def img_clahe(chain):
    """ 直方图均衡化 """
    image, show = chain
    clahe = cv2.createCLAHE(clipLimit=3.0, tileGridSize=(5, 5))
    lab = cv2.cvtColor(image, cv2.COLOR_BGR2LAB)
    l, a, b = cv2.split(lab)
    l2 = clahe.apply(l)
    lab = cv2.merge((l2, a, b))
    clahe_ = cv2.cvtColor(lab, cv2.COLOR_LAB2BGR)
    return "clahe", clahe_, show


@interceptor_util.intercept(before=_read, after=_show)
def img_remove_background(chain):
    """ 去除背景 """
    image, show = chain
    if len(image.shape) == 3:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    binary = cv2.adaptiveThreshold(image, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY_INV, 25, 15)
    se = cv2.getStructuringElement(cv2.MORPH_RECT, (1, 1))
    se = cv2.morphologyEx(se, cv2.MORPH_CLOSE, (2, 2))
    mask = cv2.dilate(binary, se)
    mask1 = cv2.bitwise_not(mask)
    binary = cv2.bitwise_and(image, mask)
    result = cv2.add(binary, mask1)
    return "remove_background", result, show


if __name__ == "__main__":
    image_file = r"C:\Users\zhousf-a\Desktop\0\07398c4c-efb238518a73537efa4b212fceaf7a6f_76_ero.jpg"
    image_file = cv2.imread(image_file, 0)
    # binary = img_binary(image_file, False)
    # dilate = img_dilate(binary, True)
    # erosion = img_erosion(dilate, True)
    # cv2.imwrite(r"C:\Users\zhousf-a\Desktop\0\3_vis_erosion.jpg", erosion)





# -*- coding:utf-8 -*-
# Author:  zhousf
# Description: coco bbox 修改工具：合并标签、删除标签
import json
from pathlib import Path


def merge_label(coco_dir: Path, labels: dict):
    """
    合并标签
    :param coco_dir: COCO目录
    :param labels: {"等级": ['等级_1', '等级_2', '等级_2E']}
    :return:
    """
    need_modify_labels = {}
    for k in labels:
        items = labels.get(k)
        for item in items:
            need_modify_labels[item] = k

    result_json_file = coco_dir.joinpath("result.json")
    with result_json_file.open("r", encoding="utf-8") as f:
        result_json = json.loads(f.read())
    # 图片
    images = result_json["images"]
    # 标签类别
    categories = result_json["categories"]
    # 标注
    annotations = result_json["annotations"]
    # info
    info = result_json["info"]
    # 遍历图片
    for img in images:
        if img["width"] is None or img["height"] is None:
            continue
        file_name = img["file_name"]
        file_name = str(file_name).rsplit("/")[-1]
        img["file_name"] = file_name
    label_dict = {}
    for item in categories:
        label_name = item.get("name")
        label_id = item.get("id")
        if label_name in need_modify_labels:
            label_name = need_modify_labels.get(label_name)
        label_dict[label_id] = label_name

    index_mapping = {}
    label_names = {}
    for index in label_dict.keys():
        name = label_dict.get(index)
        if name not in label_names:
            index_mapping[index] = index
            label_names[name] = index
        else:
            index_mapping[index] = label_names.get(name)
    print(index_mapping)
    print(label_dict)
    # 构建新的标签列表
    categories_new = {}
    for index in index_mapping:
        if index_mapping.get(index) not in categories_new:
            categories_new[index_mapping.get(index)] = label_dict.get(index)
    print(categories_new)
    categories_list = []
    for category in categories_new:
        categories_list.append({"id": category, "name": categories_new.get(category)})
    for ann in annotations:
        # 更新标签id
        print(ann.get("category_id"))
        ann["category_id"] = index_mapping.get(ann.get("category_id"))
        assert ann["category_id"] is not None
    json_file = coco_dir.joinpath("result.json")
    data_json = {
        "images": images,
        "categories": categories_list,
        "annotations": annotations,
        "info": info,
    }
    with json_file.open("w", encoding="utf-8") as f:
        json.dump(data_json, f, ensure_ascii=False, indent=4)


def delete_label(coco_dir: Path, labels: list):
    """
    删除标签
    :param coco_dir: COCO目录
    :param labels: ['等级_1', '等级_2', '等级_2E']
    :return:
    """
    result_json_file = coco_dir.joinpath("result.json")
    with result_json_file.open("r", encoding="utf-8") as f:
        result_json = json.loads(f.read())
    # 图片
    images = result_json["images"]
    # 标签类别
    categories = result_json["categories"]
    # 标注
    annotations = result_json["annotations"]
    # info
    info = result_json["info"]
    # 遍历图片
    for i in range(len(images) - 1, -1, -1):
        img = images[i]
        if img["width"] is None or img["height"] is None:
            images.pop(i)
            continue
        file_name = img["file_name"]
        file_name = str(file_name).rsplit("/")[-1]
        img["file_name"] = file_name
        print(file_name)
    delete_indexes = [item.get("id") for item in categories if item.get("name") in labels]
    for i in range(len(annotations) - 1, -1, -1):
        ann = annotations[i]
        # 删除标签
        if ann.get("category_id") in delete_indexes:
            annotations.remove(ann)
        assert ann["category_id"] is not None
    categories_list = []
    for cate in categories:
        index = cate.get("id")
        if index not in delete_indexes:
            categories_list.append(cate)
    json_file = coco_dir.joinpath("result.json")
    data_json = {
        "images": images,
        "categories": categories_list,
        "annotations": annotations,
        "info": info,
    }
    with json_file.open("w", encoding="utf-8") as f:
        json.dump(data_json, f, ensure_ascii=False, indent=4)


def update_coco(coco_dir: Path):
    result_json_file = coco_dir.joinpath("result.json")
    with result_json_file.open("r", encoding="utf-8") as f:
        result_json = json.loads(f.read())
    # 图片
    images = result_json["images"]
    # 标签类别
    categories = result_json["categories"]
    # 标注
    annotations = result_json["annotations"]
    # info
    info = result_json["info"]
    # 遍历图片
    for img in images:
        if img["width"] is None or img["height"] is None:
            continue
        file_name = img["file_name"]
        file_name = str(file_name).rsplit("/")[-1]
        img["file_name"] = file_name
    json_file = coco_dir.joinpath("result.json")
    data_json = {
        "images": images,
        "categories": categories,
        "annotations": annotations,
        "info": info,
    }
    with json_file.open("w", encoding="utf-8") as f:
        json.dump(data_json, f, ensure_ascii=False, indent=4)


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function: labelme convert coco | coco conver labelme
import json
import numpy
import shutil
import colorsys
from pathlib import Path
from zhousflib.util.img_util import get_file_base64


def coco_convert_bbox(box_coco: list):
    _x, _y, _width, _height = tuple(box_coco)
    _bbox = (_x, _y, _x + _width, _y + _height)
    return _bbox


def bbox_convert_coco(bbox: tuple):
    x_min_, y_min_, x_max_, y_max_ = bbox
    x = x_min_
    y = y_min_
    width = x_max_ - x_min_
    height = y_max_ - y_min_
    return [x, y, width, height]


def coco_convert_labelme(coco_dir: Path, dist_dir: Path, is_rectangle=False):
    """
    coco转labelme，支持rectangle和polygon
    :param coco_dir:
    :param dist_dir:
    :param is_rectangle:
    :return:
    """
    if not dist_dir.exists():
        dist_dir.mkdir()
    images_dir = coco_dir.joinpath("images")
    coco_result_file = coco_dir.joinpath("result.json")
    with coco_result_file.open("r", encoding="utf-8") as f:
        result_json = json.loads(f.read())
    # 图片
    images = result_json["images"]
    # 标签类别
    categories = result_json["categories"]
    # 标注
    annotations = result_json["annotations"]
    # info
    info = result_json["info"]

    hsv_tuples = [(1.0 * x / len(categories), 1., 1.) for x in range(len(categories))]
    colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))
    colors = list(map(lambda x: (int(x[0] * 255), int(x[1] * 255), int(x[2] * 255)), colors))

    # 遍历类别
    category_info_data = {}
    category_index = []
    for category in categories:
        category_id = category["id"]
        category_name = category["name"]
        if category_id not in category_info_data:
            category_info_data[category_id] = category_name
        if category_id not in category_index:
            category_index.append(category_id)
    # 遍历图片
    image_info_data = {}
    for i in range(len(images) - 1, -1, -1):
        img = images[i]
        if img["width"] is None or img["height"] is None:
            images.pop(i)
            continue
        file_name = img["file_name"]
        file_name = str(file_name).rsplit("/")[-1]
        img["file_name"] = file_name
        image_id = img["id"]
        if image_id not in image_info_data:
            image_info_data[image_id] = {"image_id": image_id, "file_name": file_name,
                                         "width": img["width"], "height": img["height"]}
    # 遍历标注
    image_label_data = {}
    for annotation in annotations:
        image_id = annotation["image_id"]
        if image_id in image_label_data:
            image_label_data[image_id].append(annotation)
        else:
            image_label_data[image_id] = [annotation]
    for image_id in image_label_data.keys():
        shapes = []
        image_name = image_info_data.get(image_id).get("file_name")
        for annotation in image_label_data.get(image_id):
            category_id = annotation["category_id"]
            shape = {}
            if category_id not in category_info_data:
                continue
            points = []
            shape["label"] = category_info_data.get(category_id)
            segmentation = annotation["segmentation"]
            bbox = annotation["bbox"]
            if len(segmentation) > 0:
                # todo
                continue
            if len(bbox) > 0:
                if is_rectangle:
                    _x_min, _y_min, _width, _height = tuple(bbox)
                    _x_max = _x_min + _width
                    _y_max = _y_min + _height
                    points.append([_x_min, _y_min])
                    points.append([_x_max, _y_max])
                else:
                    bbox = coco_convert_bbox(bbox)
                    x_min = int(bbox[0])
                    y_min = int(bbox[1])
                    x_max = int(bbox[2])
                    y_max = int(bbox[3])
                    points.append([x_min, y_min])
                    points.append([x_max, y_min])
                    points.append([x_max, y_max])
                    points.append([x_min, y_max])
                shape["points"] = points
                shape["line_color"] = colors[category_index.index(category_id)]
                shape["fill_color"] = ""
                shape["shape_type"] = "rectangle" if is_rectangle else "polygon"
                shape["group_id"] = ""
                shape["description"] = ""
                shape["flags"] = {}
                shapes.append(shape)
        data = {
            "flags": {},
            "shapes": shapes,
            # "lineColor": [0, 255, 0, 128],
            # "fillColor": [255, 0, 0, 128],
            "imagePath": image_name,
            "imageWidth": image_info_data.get(image_id).get("width"),
            "imageHeight": image_info_data.get(image_id).get("height"),
            "imageData": get_file_base64(images_dir.joinpath(image_name)),

        }
        shutil.copy(images_dir.joinpath(image_name), dist_dir)
        save_file = dist_dir.joinpath("{0}.json".format(image_name.split(".")[0]))
        with save_file.open("w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False, indent=4)
        print(save_file)


def labelme_convert_coco(labelme_dirs: list, dist_dir: Path):
    """
    labelme转coco， shape_type支持rectangle和polygon
    :param labelme_dirs:
    :param dist_dir:
    :return:
    """
    images_dir = dist_dir.joinpath("images")
    if not images_dir.exists():
        images_dir.mkdir(parents=True)
    images = []
    categories = []
    annotations = []
    categories_list = []
    image_id = 0
    ann_id = 0
    for labelme_dir in labelme_dirs:
        for json_file in labelme_dir.rglob("*.json"):
            print(json_file)
            with json_file.open("r", encoding="utf-8") as f:
                data = json.load(f)
                images.append({"width": data["imageWidth"],
                               "height": data["imageHeight"],
                               "file_name": data["imagePath"],
                               "id": image_id})
                image_file = json_file.parent.joinpath(data["imagePath"])
                if not image_file.exists():
                    image_file.mkdir()
                    continue
                for shape in data["shapes"]:
                    label = shape["label"]
                    if label not in categories_list:
                        categories_list.append(label)
                    points = shape["points"]
                    # shape_type 支持 rectangle 和 polygon
                    arr = numpy.asarray(points)
                    x_min = numpy.min(arr[:, 0])
                    x_max = numpy.max(arr[:, 0])
                    y_min = numpy.min(arr[:, 1])
                    y_max = numpy.max(arr[:, 1])
                    b_width = abs(x_max-x_min)
                    b_height = abs(y_max-y_min)
                    annotation = {"id": ann_id, "image_id": image_id, "category_id": categories_list.index(label),
                                  "bbox": [x_min, y_min, b_width, b_height], "segmentation": [], "ignore": 0,
                                  "iscrowd": 0, "area": b_width*b_height}
                    annotations.append(annotation)
                    ann_id += 1
                shutil.copy(image_file, images_dir)
            image_id += 1
    for i, name in enumerate(categories_list):
        categories.append({"id": i, "name": name})
    json_file = dist_dir.joinpath("result.json")
    data_json = {
        "images": images,
        "categories": categories,
        "annotations": annotations,
        "info": {}
    }
    with json_file.open("w", encoding="utf-8") as f:
        json.dump(data_json, f, ensure_ascii=False, indent=4)


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Date    : 2022/2/24 
# @Function: 非极大值抑制
import numpy as np


def nms(boxes, iou_thresh=0.8) -> list:
    """
    仅适合单个类别，不适合多类别
    :param boxes: [[scores, x_min, y_min, x_max, y_max]....]
    :param iou_thresh: 交并比的阈值，当大于该阈值时则选优
    :return: [index] boxes的索引
    """
    if isinstance(boxes, list):
        boxes = np.array(boxes)
    x1 = boxes[:, -4].astype(float).astype(int)
    y1 = boxes[:, -3].astype(float).astype(int)
    x2 = boxes[:, -2].astype(float).astype(int)
    y2 = boxes[:, -1].astype(float).astype(int)
    areas = (y2 - y1 + 1) * (x2 - x1 + 1)
    scores = boxes[:, -5].astype(float)
    # 存放nms后剩余的box
    boxes_filter = []
    # 取出分数从大到小排列的索引 .argsort()是从小到大排列，[::-1]是列表头和尾颠倒一下
    index = scores.argsort()[::-1]
    # print('index:',index)
    # 上面这两句比如分数[0.72 0.8  0.92 0.72 0.81 0.9 ]
    #  对应的索引index[  2   5    4     1    3   0  ]
    #  记住是取出索引，scores列表没变
    # index会剔除遍历过的方框，和合并过的box。
    while index.size > 0:
        # print(index.size)
        # 取出第一个方框进行和其他方框比对，看有没有可以合并的
        i = index[0]  # 取出第一个索引号，这里第一个是【2】

        # 因为我们这边分数已经按从大到小排列了。
        # 所以如果有合并存在，也是保留分数最高的这个，也就是我们现在那个这个
        # keep保留的是索引值，不是具体的分数。
        boxes_filter.append(i)
        # print('keep:',keep)
        # print('x1:', x1[i])
        # print(x1[index[1:]])

        # 计算交集的左上角和右下角
        # 这里要注意，比如x1[i]这个方框的左上角x和所有其他的方框的左上角x的
        x11 = np.maximum(x1[i], x1[index[1:]])  # calculate the points of overlap
        y11 = np.maximum(y1[i], y1[index[1:]])
        x22 = np.minimum(x2[i], x2[index[1:]])
        y22 = np.minimum(y2[i], y2[index[1:]])
        # print('*'*20)
        # print(x11, y11, x22, y22)
        # 这边要注意，如果两个方框相交，X22-X11和Y22-Y11是正的。
        # 如果两个方框不相交，X22-X11和Y22-Y11是负的，我们把不相交的W和H设为0.
        w = np.maximum(0, x22 - x11 + 1)
        h = np.maximum(0, y22 - y11 + 1)

        # 计算重叠面积就是上面说的交集面积。不相交因为W和H都是0，所以不相交面积为0
        overlaps = w * h
        # print('overlaps is', overlaps)

        # 这个就是IOU公式（交并比）。
        # 得出来的ious是一个列表，里面拥有当前方框和其他所有方框的IOU结果。
        ious = overlaps / (areas[i] + areas[index[1:]] - overlaps)
        # print('ious is', ious)
        # print(type(ious))

        # 接下来是合并重叠度最大的方框，也就是合并ious中值大于thresh的方框
        # 我们合并的操作就是把他们剔除，因为我们合并这些方框只保留下分数最高的。
        # 我们经过排序当前我们操作的方框就是分数最高的，所以我们剔除其他和当前重叠度最高的方框
        # 这里np.where(ious<=thresh)[0]是一个固定写法。
        idx = np.where(ious <= iou_thresh)[0]
        # print('idx:',idx)
        # print(type(idx))

        # 把留下来框在进行NMS操作
        # 这边留下的框是去除当前操作的框，和当前操作的框重叠度大于thresh的框
        # 每一次都会先去除当前操作框，所以索引的列表就会向前移动移位，要还原就+1，向后移动一位
        index = index[idx + 1]  # because index start from 1
    return boxes_filter


def multi_nms(boxes: list, iou_thresh=None) -> list:
    """
    适合多类别
    :param boxes: [[cls, scores, x_min, y_min, x_max, y_max]....]
    :param iou_thresh: 交并比的阈值，当大于该阈值时则选优
        {   "base": 0.8,
            "cls_1": 0.5,
            "cls_2": 0.5
        }
    :return: [index] boxes的索引
    """
    if iou_thresh is None:
        iou_thresh = {"base": 0.8}
    if isinstance(boxes, list):
        boxes = np.array(boxes)
    classes = boxes[:, -6]
    res_ids = []
    unique_cls = np.unique(classes)
    # 遍历每个类别，进行nms操作
    for cls in unique_cls:
        det = [box for box in boxes if box[0] == cls]
        det_ids = [index for index, box in enumerate(boxes) if box[0] == cls]
        thresh = iou_thresh.get(cls, iou_thresh.get("base"))
        res = nms(boxes=det, iou_thresh=thresh)
        if len(res):
            for i in res:
                res_ids.append(det_ids[i])
    return res_ids


if __name__ == "__main__":
    # boxes_ = np.array([[0.72, 100, 100, 210, 210],
    #                   [0.80, 250, 250, 420, 420],
    #                   [0.92, 220, 220, 320, 330],
    #                   [0.72, 100, 100, 210, 210],
    #                   [0.81, 230, 240, 325, 330],
    #                   [0.90, 220, 230, 315, 340]])
    # print(nms(boxes_, iou_thresh=0.7))
    boxes_ = np.array([["1", 0.72, 100, 100, 210, 210],
                      ["1", 0.80, 250, 250, 420, 420],
                      ["1", 0.92, 220, 220, 320, 330],
                      ["2", 0.72, 100, 100, 210, 210],
                      ["2", 0.81, 230, 240, 325, 330],
                      ["2", 0.90, 220, 230, 315, 340]]).tolist()
    print(multi_nms(boxes_, iou_thresh={"base": 0.5, "1": 0.7, "2": 0.9}))




# -*- coding:utf-8 -*-
# Author:  zhousf
# Description: coco bbox 提取工具
import json
from pathlib import Path
from PIL import Image, ImageDraw


def compute_contain(box1, box2):
    """
    计算两个box是否为包含关系
    :param box1: (x_min, y_min, x_max, y_max)
    :param box2: (x_min, y_min, x_max, y_max)
    :return: 返回两个box重叠面积占较小box的面积比，一般大于0.8则为包含关系
    box1=(317,280,553,395)
    box2=(374,295,485,322)
    """
    px_min, py_min, px_max, py_max = box1
    gx_min, gy_min, gx_max, gy_max = box2
    p_area = (px_max - px_min) * (py_max - py_min)  # 计算P的面积
    g_area = (gx_max - gx_min) * (gy_max - gy_min)  # 计算G的面积
    # 求相交矩形的左下和右上顶点坐标(x_min, y_min, x_max, y_max)
    _x_min = max(px_min, gx_min)  # 得到左下顶点的横坐标
    _y_min = max(py_min, gy_min)  # 得到左下顶点的纵坐标
    _x_max = min(px_max, gx_max)  # 得到右上顶点的横坐标
    _y_max = min(py_max, gy_max)  # 得到右上顶点的纵坐标
    # 计算相交矩形的面积
    w = _x_max - _x_min
    h = _y_max - _y_min
    if w <= 0 or h <= 0:
        return 0
    area = w * h  # G∩P的面积
    if p_area >= g_area:
        return area / g_area
    else:
        return area / p_area


def coco_convert_bbox(box_coco: list):
    _x, _y, _width, _height = tuple(box_coco)
    _bbox = (_x, _y, _x + _width, _y + _height)
    return _bbox


def bbox_convert_coco(bbox: tuple):
    x_min_, y_min_, x_max_, y_max_ = bbox
    x = x_min_
    y = y_min_
    width = x_max_ - x_min_
    height = y_max_ - y_min_
    return [x, y, width, height]


def box_expand(box, offset=50):
    x_min, y_min, x_max, y_max = box
    offset = min(x_min, y_min) if min(x_min, y_min) < offset else offset
    _x_min = x_min - offset
    _x_max = x_max + offset
    _y_min = y_min - offset
    _y_max = y_max + offset
    return _x_min, _y_min, _x_max, _y_max


def generate_image_by_label(data_dirs: list, dist_dir: Path, labels: list, contain_child: bool = False):
    """
    根据标签生成标注数据，进行图片裁剪
    :param data_dirs:
    :param dist_dir:
    :param labels: 提取标签名称
    :param contain_child: 提取时包括区域内的所有标签
    :return:
    """
    dist_dir = dist_dir.joinpath("images")
    if not dist_dir.exists():
        dist_dir.mkdir(parents=True)
    anno_new = []
    img_new = []
    categories_new = []
    info_new = {}
    img_index = 0
    anno_id = 0
    for data_dir in data_dirs:
        if not data_dir.is_dir():
            continue
        for label_json in data_dir.rglob("*.json"):
            images = {}
            annotations = {}
            with label_json.open("r", encoding="utf-8") as f:
                result_json = json.loads(f.read())
            # 图片
            _images = result_json["images"]
            # 标签类别
            _categories = result_json["categories"]
            if not contain_child:
                for cate in _categories:
                    if cate.get("name") in labels:
                        categories_new.append(cate)
            else:
                categories_new = _categories
            if len(categories_new) == 0:
                print("label not exist: {0}".format(labels))
                return
            label_dict = {item.get("id"): item.get("name") for item in _categories}
            # 标注
            _annotations = result_json["annotations"]
            # info
            _info = result_json["info"]
            info_new = _info
            # 遍历图片
            for img in _images:
                file_name = img["file_name"]
                file_name = str(file_name).rsplit("/")[-1]
                images[img["id"]] = data_dir.joinpath("images").joinpath(file_name)
            # 遍历标签
            for anno in _annotations:
                image_id = anno["image_id"]
                anno["id"] = anno_id
                anno_id += 1
                if image_id not in annotations:
                    annotations[image_id] = [anno]
                else:
                    annotations[image_id].append(anno)
                assert anno["category_id"] is not None
            for img_id in annotations:
                anno_list = annotations.get(img_id)
                steel_graph_anno_list = []
                others_anno = []
                for anno in anno_list:
                    category_id = anno["category_id"]
                    category_name = label_dict.get(category_id)
                    if category_name in labels:
                        steel_graph_anno_list.append(anno)
                    else:
                        others_anno.append(anno)
                img = Image.open(images.get(img_id))
                if img.mode != "RGB":
                    img = img.convert('RGB')
                if len(steel_graph_anno_list) > 0 and len(others_anno) > 0:
                    for steel_graph_anno in steel_graph_anno_list:
                        contain_anno = []
                        bbox = coco_convert_bbox(steel_graph_anno["bbox"])
                        for i in range(len(others_anno) - 1, -1, -1):
                            _anno = others_anno[i].copy()
                            _bbox = coco_convert_bbox(_anno["bbox"])
                            iou = compute_contain(bbox, _bbox)
                            if iou > 0.5:
                                contain_anno.append(_anno)
                        if not steel_graph_anno:
                            continue
                        bbox = coco_convert_bbox(steel_graph_anno["bbox"])
                        bbox_big = box_expand(box=bbox, offset=10)
                        # 裁剪图片
                        cropped = img.crop(bbox_big)
                        x_min = bbox[0] - bbox_big[0]
                        y_min = bbox[1] - bbox_big[1]
                        x_max = bbox[2] - bbox_big[0]
                        y_max = bbox[3] - bbox_big[1]
                        steel_graph_anno["bbox"] = bbox_convert_coco((x_min,  y_min, x_max, y_max))
                        steel_graph_anno["image_id"] = img_index
                        anno_new.append(steel_graph_anno)
                        # draw = ImageDraw.ImageDraw(cropped)
                        # draw.rectangle(xy=(x_min,  y_min, x_max, y_max), fill=None, outline="red", width=1)
                        for anno in contain_anno:
                            _bbox = coco_convert_bbox(anno["bbox"])
                            # 坐标偏移计算
                            x_min_ = _bbox[0] - bbox_big[0]
                            y_min_ = _bbox[1] - bbox_big[1]
                            x_max_ = _bbox[2] - bbox_big[0]
                            y_max_ = _bbox[3] - bbox_big[1]
                            anno["bbox"] = bbox_convert_coco((x_min_,  y_min_, x_max_, y_max_))
                            anno["image_id"] = img_index
                            # draw.rectangle(xy=(x_min_,  y_min_, x_max_, y_max_), fill=None, outline="red", width=1)
                        # cropped.show()
                        if contain_child:
                            anno_new.extend(contain_anno)
                        save_img_file = dist_dir.joinpath("{0}_{1}{2}".format(images.get(img_id).stem,img_index, images.get(img_id).suffix))
                        cropped.save(save_img_file)
                        img_new.append({"width": cropped.width, "height": cropped.height, "id": img_index,
                                        "file_name": save_img_file.name})
                        img_index += 1
                        print(img_index)
    result_file = dist_dir.parent.joinpath("result.json")
    with open(result_file, 'w', encoding="utf-8") as f:
        json.dump({
            "images": img_new,
            "categories": categories_new,
            "annotations": anno_new,
            "info": info_new
        }, f, ensure_ascii=False, indent=4)


if __name__ == "__main__":
    pass





# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function: 日志工具
import os
import time
import codecs
import logging.handlers
from logging.handlers import BaseRotatingHandler
from flask.logging import default_handler
from flask import current_app


# noinspection PyPep8Naming
class MultiProcessSafeDailyRotatingFileHandler(BaseRotatingHandler):
    """Similar with `logging.TimedRotatingFileHandler`, while this one is
    - Multi process safe
    - Rotate at midnight only
    - Utc not supported
    日志handler：按每天分割
    """

    def __init__(self, filename, encoding=None, delay=False, utc=False, **kwargs):
        self.utc = utc
        self.suffix = "%Y-%m-%d.txt"
        self.baseFilename = filename
        self.currentFileName = self._compute_fn()
        BaseRotatingHandler.__init__(self, filename, 'a', encoding, delay)

    def shouldRollover(self, record):
        if self.currentFileName != self._compute_fn():
            return True
        return False

    def doRollover(self):
        if self.stream:
            self.stream.close()
            self.stream = None
        self.currentFileName = self._compute_fn()

    def _compute_fn(self):
        return self.baseFilename + "-" + time.strftime(self.suffix, time.localtime())

    def _open(self):
        if self.encoding is None:
            stream = open(self.currentFileName, self.mode)
        else:
            stream = codecs.open(self.currentFileName, self.mode, self.encoding)
        # simulate file name structure of `logging.TimedRotatingFileHandler`
        if os.path.exists(self.baseFilename):
            try:
                os.remove(self.baseFilename)
            except OSError as e:
                print(e)
        try:
            os.symlink(self.currentFileName, self.baseFilename)
        except OSError as e:
            print(e)
        return stream


"""
日志级别
CRITICAL = 50
FATAL = CRITICAL
ERROR = 40
WARNING = 30
WARN = WARNING
INFO = 20
DEBUG = 10
NOTSET = 0
"""


def init_log(app, log_dir):
    app.logger.removeHandler(default_handler)
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)
    log_file_str = os.path.join(log_dir, 'log')
    time_handler = MultiProcessSafeDailyRotatingFileHandler(log_file_str, encoding='utf-8')
    logging_format = logging.Formatter(
        '=>%(asctime)s - [%(levelname)s]\n%(message)s')
    # logging_format = logging.Formatter(
    #     "%(asctime)s-[%(threadName)s-%(thread)d]-%(levelname)s-[%(filename)s:%(lineno)d]\n%(message)s")
    logging.basicConfig(level=logging.DEBUG)
    time_handler.setFormatter(logging_format)
    app.logger.addHandler(time_handler)


def error(msg, *args, **kwargs):
    current_app.logger.error(msg, *args, **kwargs)


def warning(msg, *args, **kwargs):
    current_app.logger.warning(msg, *args, **kwargs)


def info(msg, *args, **kwargs):
    current_app.logger.info(msg, *args, **kwargs)


def debug(msg, *args, **kwargs):
    current_app.logger.debug(msg, *args, **kwargs)





# -*- coding:utf-8 -*-
# Author:      zhousf
# File:        permission_util.py
# Description: 动态权限管理
import os
import json


class DynamicPermission(object):

    def __init__(self, visit_file):
        """

        :param visit_file: json file
        """
        self.visit_file = visit_file
        if self.visit_file is None or not os.path.exists(self.visit_file):
            raise Exception("visit_file is not exist.")
        self.visit_list = self.load()

    def load(self):
        return self.__load_dict_from_json_file(self.visit_file)

    def update(self):
        self.visit_list = self.load()

    def check_visit(self, visit):
        if visit in self.visit_list.keys():
            return True
        tmp = self.load()
        if visit in tmp.keys():
            self.update()
            return True
        return False

    def fetch_api_secret(self, req_code):
        return self.visit_list.get(req_code)

    @staticmethod
    def __load_dict_from_json_file(json_file):
        """
        json文件转成字典
        :param json_file:
        :return: dict
        """
        with open(json_file, 'r') as f:
            content = f.read()
            if content is not None and content.strip() != '':
                return json.loads(s=content)
        return None



# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function: lmdb (Lightning Memory-Mapped Database) 快如闪电的内存映射数据库
# pip install lmdb
import pickle
import lmdb
from pathlib import Path

"""
实例化数据库
db = LMDB(Path(__file__).parent)
插入数据
db.insert("student", {"a": 99, "b": 95})
删除数据
db.delete("student")
修改数据
db.update("student", {"a": 98, "b": 100})
查询数据
print(db.query("student"))
显示所有数据
db.display()
"""


class LMDB(object):

    def __init__(self, db_dir: Path):
        """
        初始化
        :param db_dir: 文件目录
        """
        self.env = self.initialize(db_dir)

    @staticmethod
    def initialize(db_dir: Path):
        return lmdb.open(path=str(db_dir), map_size=int(1e9))

    def insert(self, key: str, value):
        """
        插入数据
        :param key:
        :param value:
        :return:
        """
        txn = self.env.begin(write=True)
        txn.put(key.encode(), pickle.dumps(value))
        txn.commit()

    def delete(self, key: str):
        """
        删除
        :param key:
        :return:
        """
        txn = self.env.begin(write=True)
        txn.delete(key.encode())
        txn.commit()

    def update(self, key: str, value):
        """
        更新
        :param key:
        :param value:
        :return:
        """
        self.insert(key, value)

    def query(self, key: str):
        """
        查询
        :param key:
        :return:
        """
        txn = self.env.begin()
        res = txn.get(key.encode())
        return None if res is None else pickle.loads(res)

    def display(self):
        """
        显示所有数据
        :return:
        """
        txn = self.env.begin()
        cur = txn.cursor()
        for key, value in cur:
            print(key, pickle.loads(value))

    def clear_all(self):
        """
        删除所有数据
        :return:
        """
        txn = self.env.begin()
        cur = txn.cursor()
        for key, value in cur:
            print(key, pickle.loads(value))
            txn.delete(key.encode())
        txn.commit()

    def close(self):
        self.env.close()


if __name__ == '__main__':
    pass




# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Date    : 2023/10/18 
# @Function:
import json
import numpy
import shutil
from pathlib import Path


def rectangle_convert_polygon(labelme_dir: Path, dst_dir: Path):
    """

    :param labelme_dir:
    :param dst_dir:
    :return:
    """
    if not dst_dir.exists():
        dst_dir.mkdir()
    for label_file in labelme_dir.glob('*.json'):
        print(label_file)
        with label_file.open(mode="r", encoding="utf-8") as f:
            data = json.load(f)
            imagePath = data["imagePath"]
            shapes = data['shapes']
            for i in range(0, len(shapes)):
                points = shapes[i]['points']
                shape_type = shapes[i]['shape_type']
                if shape_type == "rectangle":
                    arr = numpy.asarray(points)
                    x_min = numpy.min(arr[:, 0])
                    x_max = numpy.max(arr[:, 0])
                    y_min = numpy.min(arr[:, 1])
                    y_max = numpy.max(arr[:, 1])
                    polygon = [[x_min, y_min], [x_max, y_min], [x_max, y_max], [x_min, y_max]]
                    shapes[i]['points'] = polygon
                    shapes[i]['shape_type'] = "polygon"
            data['shapes'] = shapes
        save_json = dst_dir.joinpath(label_file.name)
        save_image = label_file.parent.joinpath(imagePath)
        if not save_image.exists():
            continue
        with save_json.open("w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False, indent=4)
        shutil.copy(save_image, dst_dir)


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Date    : 2021/8/9 
# @Function: web服务配置与管理
import os
import time
import random
from flask import g
from app import FLASK_APP
from gevent import monkey
from zhousflib.web import log_util
from flask import request, Response
from gevent.pywsgi import WSGIServer
from zhousflib.web import response as res
from werkzeug.exceptions import HTTPException


class WebApp(object):
    def __init__(self, config_app):
        FLASK_APP.config.from_object(config_app)
        self.config_app = config_app
        self.app = FLASK_APP
        log_util.init_log(FLASK_APP, config_app.LOG_SERVICE_DIR)
        self.init_config()

    def save_pid(self):
        if not self.config_app.has_pid_file(self.config_app):
            return None
        pid = os.getpid()
        with open(file=self.config_app.PID_FILE, mode='a+') as f:
            f.write(pid.__str__() + "\n")
        return pid

    def start(self):
        server = WSGIServer((self.config_app.HOST, self.config_app.PORT), self.app)
        print("http://{0}:{1}".format(self.config_app.HOST, self.config_app.PORT))
        # pid = self.save_pid()
        # print("pid: {0}".format(pid))
        print("{0} app is starting...".format(self.config_app.ENVIRONMENT))
        try:
            server.serve_forever()
        except Exception as e:
            print(e)
            os.remove(self.config_app.PID_FILE)

    def init_config(self):
        print("create app...")
        if not self.config_app.DEBUG:
            monkey.patch_all()
        FLASK_APP.jinja_env.auto_reload = True
        FLASK_APP.config['MAX_CONTENT_LENGTH'] = 30 * 1024 * 1024
        FLASK_APP.json.ensure_ascii = self.config_app.JSON_AS_ASCII
        FLASK_APP.json.sort_keys = self.config_app.JSON_SORT_KEYS
        FLASK_APP.json.mimetype = self.config_app.JSONIFY_MIMETYPE

        @FLASK_APP.before_request
        def before_request():
            request_id = '{0}_{1}'.format(int(round(time.time() * 1000)), self.rand(digits=5))
            req_id = request.values.get("reqId", None)
            if req_id:
                request_id = req_id
            real_ip = request.headers.get('X-Real-Ip', request.remote_addr)
            g.request_id = request_id
            g.real_ip = real_ip
            g.request_time = time.time()
            g.static = False
            g.log = ""
            msg = time.strftime('%Y/%m/%d %H:%M:%S', time.localtime(time.time())) + '\n'
            msg += '[request:{0} | ip:{1} | url:{2} | method:{3}]\n'.format(request_id, real_ip, request.path, request.method)
            if request.view_args is not None and len(request.view_args) > 0:
                g.static = True
                delattr(g, "log")
                return
            params = request.data
            if params is None:
                g.static = True
                delattr(g, "log")
                return
            params = bytes.decode(params, encoding='utf-8')
            params = params.strip()
            if 0 < len(params) < 10240:
                msg += params + "\n"
            g.log = msg

        @FLASK_APP.after_request
        def after_request(response):
            """
            RuntimeError: Attempted implicit sequence conversion but
                        the response object is in direct passthrough mode.
            response.direct_passthrough = False
            """
            response.direct_passthrough = False
            msg_before = ""
            if hasattr(g, "log"):
                msg_before += g.log
            if hasattr(g, "logger"):
                g.logger.title_first("返回信息")
            cost = round(time.time() - g.request_time, 4)
            msg_after = '[response:{0} | cost_time:{1}s]\n'.format(g.request_id, cost)
            status_code = 200
            if isinstance(response, Response) and not g.static:
                if response.json is not None and response.content_length < 10240:
                    msg_after += "{0}\n".format(response.json)
                    status_code = response.json.get("code", status_code)
                if hasattr(g, "status_code"):
                    status_code = g.status_code
                if hasattr(g, "logger"):
                    if status_code in range(201, 500):
                        g.logger.log(msg_after)
                        log_util.warning(msg_before + msg_after)
                    elif status_code in range(500, 700):
                        g.logger.log(msg_after)
                        log_util.error(msg_before + msg_after)
                    elif status_code > 1000:
                        g.logger.log_txt = g.logger.log_txt.replace("\n\n", "\n")
                        g.logger.log(msg_after)
                        log_util.error(g.logger.log_txt)
                    else:
                        g.logger.log(msg_after)
                        log_util.info(msg_before + msg_after)
                    g.logger.save_log()
                else:
                    if hasattr(g, "log"):
                        log_util.error(msg_before + msg_after)
            return response

        @FLASK_APP.route('/')
        def index():
            ip, port = request.host.split(":")
            return res.success_tip(status=200, result="{0}:{1}".format(ip, port))

        @FLASK_APP.route('/favicon.ico')
        def get_fav():
            return FLASK_APP.send_static_file('images/favicon.ico')

        @FLASK_APP.errorhandler(HTTPException)
        def exception(e):
            log_util.warning(repr(e))
            g.status_code = e.code
            return res.failed_tip(status=e.code, result="{0}[{1}]".format(e.description, request.path))

    @staticmethod
    def rand(digits=3):
        """
        生成随机数
        :param digits: 随机数位数
        digits=3则生成范围[100,999]
        digits=4则生成范围[1000,9999]
        :return:
        """
        if digits < 1:
            return 0
        assert isinstance(digits, int)
        start = pow(10, digits - 1)
        end = pow(10, digits) - 1
        return random.randint(start, end)



# -*- coding:utf-8 -*-
# Author:      zhousf
# Description:  随机数
import random


def rand(digits=3):
    """
    生成随机数
    :param digits: 随机数位数
    digits=3则生成范围[100,999]
    digits=4则生成范围[1000,9999]
    :return:
    """
    if digits < 1:
        return 0
    assert isinstance(digits, int)
    start = pow(10, digits - 1)
    end = pow(10, digits) - 1
    return random.randint(start, end)


if __name__ == '__main__':
    print(rand(digits=3))
    pass



# -*- coding:utf-8 -*-
# Author:  zhousf
# Description:
import imghdr
import base64
import hashlib
from pathlib import Path


def get_file_base64(file_path: Path, contain_file_name=False, split_char=","):
    """
    图片转base64
    :param file_path: 图片路径
    :param contain_file_name: 是否包含文件名称
    :param split_char: 分隔符
    :return: 'a.jpg,iVBORw0KGgoAAAANSUhEUgAABNcAAANtCAYAAACzHZ25AAA.....'
    """
    with file_path.open('rb') as infile:
        s = infile.read()
    base64_str = base64.b64encode(s).decode("utf-8")
    if contain_file_name:
        base64_str = file_path.name + split_char + base64_str
    return base64_str


def md5(file_path: Path):
    with file_path.open('rb') as f:
        return hashlib.md5(f.read()).hexdigest()


def rename_image_with_md5(src_dir: Path, dst_dir: Path):
    if not dst_dir.exists():
        dst_dir.mkdir()
    count = 0
    repeat = 0
    for file in src_dir.rglob("*.*"):
        if not imghdr.what(str(file)):
            continue
        print(file.name)
        count += 1
        new_name = md5(file)
        new_name += file.suffix
        print(new_name)
        if dst_dir.joinpath(new_name).exists():
            repeat += 1
            continue
        if not dst_dir.joinpath(file.parent.name).exists():
            dst_dir.joinpath(file.parent.name).mkdir(parents=True)
        if not dst_dir.joinpath(file.parent.name).joinpath(new_name).exists():
            file.rename(dst_dir.joinpath(file.parent.name).joinpath(new_name))
    print("count=", count)
    print("repeat=", repeat)


if __name__ == "__main__":
    pass



# -*- coding:utf-8 -*-
# Author:  zhousf
# Description:
from concurrent.futures import ThreadPoolExecutor
import time
import threading


class ThreadPool(object):

    def __init__(self, max_workers=None):
        self.thread_pool = ThreadPoolExecutor(max_workers)
        pass

    def push(self,  fn, args):
        return self.thread_pool.submit(fn, args)


def task(x):
    print('{0} 参数: {1}'.format(threading.current_thread().getName(), x))
    time.sleep(2)
    return x


if __name__ == '__main__':
    start = time.time()
    tp = ThreadPool(max_workers=5)
    obj_list = []
    for i in range(10):
        obj_list.append(tp.push(task, i))
    print('主线程')
    for i in obj_list:
        print(i.result())
    print("耗时：{0}秒".format(round(time.time() - start, 4)))




# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function: labelme标注数据裁剪工具
import json
import numpy as np
from pathlib import Path
from PIL import Image, ImageDraw, ImageFont
from zhousflib.util.img_util import get_file_base64
from zhousflib.font import Font_SimSun
FONT = ImageFont.truetype(font=str(Font_SimSun), size=15)


def clip_img(labelme_dir: Path, dst_dir: Path, clip_labels: list, show=False):
    """
    裁剪
    :param labelme_dir:
    :param dst_dir:
    :param clip_labels: 裁剪标签
    :param show:
    :return:
    """
    if not dst_dir.exists():
        dst_dir.mkdir()
    for json_file in labelme_dir.glob("*.json"):
        print(json_file)
        with json_file.open("r", encoding="utf-8") as f:
            data = json.load(f)
        imagePath = data["imagePath"]
        image_file = labelme_dir.joinpath(imagePath)
        image = Image.open(image_file)
        if image.mode != "RGB":
            image = image.convert('RGB')
        image_w, image_h = image.size
        draw = ImageDraw.ImageDraw(image)
        for i in range(0, len(data["shapes"])):
            shape = data["shapes"][i]
            label = shape.get("label", None)
            points = shape.get("points", None)
            if not label or not points:
                continue
            if label not in clip_labels:
                continue
            p_arr = np.asarray(points)
            scale_up_pixel = 5
            x_min = np.min(p_arr[:, 0]) - scale_up_pixel
            x_max = np.max(p_arr[:, 0]) + scale_up_pixel
            y_min = np.min(p_arr[:, 1]) - scale_up_pixel
            y_max = np.max(p_arr[:, 1]) + scale_up_pixel
            y_min = y_min if y_min > 0 else 1
            x_min = x_min if x_min > 0 else 1
            x_max = x_max if x_max < image_w else image_w - 1
            y_max = y_max if y_max < image_h else image_h - 1
            # 裁剪的坐标原点换算
            origin_point = (x_min, y_min)
            p_arr_1 = p_arr[:, 0] - x_min
            p_arr_2 = p_arr[:, 1] - y_min
            p_arr_clip = np.stack([p_arr_1, p_arr_2], 1).tolist()
            # 裁剪
            cropped = image.crop((x_min, y_min, x_max, y_max))
            save_img_file = dst_dir.joinpath("{0}_{1}{2}".format(image_file.stem, i, image_file.suffix))
            cropped.save(save_img_file, quality=100)
            image_w_clip, image_h_clip = cropped.size
            # 写json文件
            save_file = dst_dir.joinpath("{0}_{1}.json".format(image_file.stem, i))
            data_clip = {"imagePath": str(save_img_file.name),
                         "imageData": get_file_base64(save_img_file),
                         "imageHeight": image_h_clip,
                         "imageWidth": image_w_clip}
            shape_clip = shape.copy()
            shape_clip["points"] = p_arr_clip
            data_clip["shapes"] = [shape_clip]
            if "version" in data:
                data_clip["version"] = data["version"]
            if "flags" in data:
                data_clip["flags"] = data["flags"]
            with save_file.open("w", encoding="utf-8") as f:
                json.dump(data_clip, f, ensure_ascii=False, indent=4)
            if show:
                width = abs(x_max - x_min)
                height = abs(y_max - y_min)
                draw.rectangle(xy=(x_min, y_min, x_min + width, y_min + height), fill=None, outline="red", width=1)
                fw, fh = FONT.getsize(label)
                if y_min < fh:
                    y_min = y_min + fh
                if (x_min + fw) > image_w:
                    x_min = x_max - fw
                draw.text(xy=(x_min, y_min - fh), text=label, fill="red", font=FONT)
        if show:
            image.show()
            break


if __name__ == "__main__":
    pass







# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function: bbox 修改工具：合并标签、删除标签
import json
from pathlib import Path


def merge_label(labelme_dir: Path, labels: dict):
    """
    todo 合并标签
    :param labelme_dir: labelme目录
    :param labels: {"等级": ['等级_1', '等级_2', '等级_2E']}
    :return:
    """
    need_modify_labels = {}
    for k in labels:
        items = labels.get(k)
        for item in items:
            need_modify_labels[item] = k


def delete_label(labelme_dir: Path, labels: list):
    """
    删除标签
    :param labelme_dir: labelme目录
    :param labels: ['等级_1', '等级_2', '等级_2E']
    :return:
    """
    for json_file in labelme_dir.rglob("*.json"):
        print(json_file)
        with json_file.open("r", encoding="utf-8") as f:
            data = json.load(f)
            for i in range(len(data["shapes"]) - 1, -1, -1):
                shape = data["shapes"][i]
                label = shape["label"]
                if label in labels:
                    data["shapes"].pop(i)
        with json_file.open("w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False, indent=4)


if __name__ == "__main__":
    pass



# -*- coding:utf-8 -*-
# Author:      zhousf
# File:        pandas_util.py
# Description:
# pip install xlrd==1.2.0  -i http://mirrors.aliyun.com/pypi/simple/ --trusted-host mirrors.aliyun.com
# pip install openpyxl  -i http://mirrors.aliyun.com/pypi/simple/ --trusted-host mirrors.aliyun.com
# pip install pandas  -i http://mirrors.aliyun.com/pypi/simple/ --trusted-host mirrors.aliyun.com
# pip install pymars -i http://mirrors.aliyun.com/pypi/simple/ --trusted-host mirrors.aliyun.com
import os
import csv
import pandas as pd
from pathlib import Path
import mars.dataframe as md
from zhousflib.pandas import openpyxl_util


def read_csv(file_path: str, header="infer", title=None, encoding=None, nrows=None, dtype=None, sep="\t"):
    """
    取值：
        csv_data['column_name']
        csv_data['column_name'].values
        columns = df_data.columns.values.tolist()
        data = df_data.values.tolist()
    :param file_path:
    :param header: 当None则不返回列名
    :param title:
    :param encoding: gbk/utf-8
    :param nrows: 读取的行数
    :param dtype: 指定数据类型
    :param sep: 分隔标志
    :return:
    """
    if not os.path.exists(file_path):
        raise Exception("file not exists: {}".format(file_path))
    if header:
        return pd.read_csv(file_path, header=header, usecols=title, encoding=encoding, nrows=nrows, dtype=dtype, sep=sep)
    else:
        return pd.read_csv(file_path, usecols=title, encoding=encoding, nrows=nrows, dtype=dtype, sep=sep)


def read_csv_mars(csv_file):
    """
    大文件读取，采用mars
    Mars 是基于张量的，用于进行大规模数据计算的统一计算框架
    :param csv_file:
    :return:
    """
    if not os.path.exists(csv_file):
        raise Exception("file not exists: {}".format(csv_file))
    return md.read_csv(csv_file, low_memory=False).execute().fetch()


def write_csv(file_path: Path, data, columns=None, seq=None):
    """
    按列写入csv
    :param file_path: '/home/data.csv'
    :param data:
        当 columns=None时
        data = {"周一": ['语文', '英语', '物理', '数学', '化学']，
                "周二": ['音乐', '英语', '数学', '地理', '语文']}
        当 columns=["周一", "周二"]
        data = [['语文', '英语', '物理', '数学', '化学'], ['音乐', '英语', '数学', '地理', '语文']]
    :param columns:
    :param seq:
    :return:
    """
    data_frame = pd.DataFrame(data)
    header = True if columns else False
    data_frame.to_csv(file_path, header=header, columns=columns, index=False, sep=seq)


def read_excel(file_path, sheet_name=None, header=None):
    """
    读取excel文件
    :param file_path:
    :param sheet_name: None第一张表
    :param header: 为None则无表头，为0则第一行为表头
    :return:
    """
    if sheet_name is None:
        exc = pd.ExcelFile(file_path)
        sheets = exc.sheet_names
        if len(sheets) > 0:
            sheet_name = sheets[0]
    data_ = pd.read_excel(file_path, sheet_name=sheet_name, dtype=object, header=header)
    # 若nan则替换成空字符串
    data_ = data_.fillna("")
    return data_


def read_excel_merge_cell(file_path: Path, sheet_name=None, delete_duplicates_rate: float = 1.0,
                          tmp_excel: Path = None, header=None):
    """
    读取excel文件，并处理合并单元格
    :param file_path: excel文件
    :param sheet_name: None第一张表
    :param delete_duplicates_rate: 对拆分合并单元格的结果去重的比例，默认为1.0（全相同时去重）,0则不去重
    :param tmp_excel: 临时文件，若为空则会更新源文件，合并单元格选项
    :param header:
    :return:
    """
    excel_file = openpyxl_util.unmerge_and_fill_cells(excel_file=file_path, target_sheet_name=sheet_name,
                                                      delete_duplicates_rate=delete_duplicates_rate, tmp_excel=tmp_excel)
    return read_excel(str(excel_file), sheet_name=sheet_name, header=header)


def write_excel(data, columns=None, save_file: Path = Path('output.xlsx'), sheet='Sheet1'):
    """
    写入excel表格
    :param data: [[1, 1], [2, 2]]
    :param columns: ['col1', 'col2']
    :param save_file:
    :param sheet:
    :return:
    """
    writer = pd.ExcelWriter(save_file)
    df1 = pd.DataFrame(data=data, columns=columns)
    df1.to_excel(writer, sheet, index=False)
    writer.close()


def print_shape(file_path):
    """
    打印数据行列数量
    :param file_path:
    :return: (1377615, 330)
    """
    if file_path.endswith("csv"):
        data = pd.read_csv(file_path, sep='\t')
    else:
        data = pd.read_excel(file_path)
    print(data.shape)


def fetch_row_csv(csv_file, save_csv, rows, reverse=False):
    """
    读取数据并保存到save_csv中
    :param csv_file:
    :param save_csv:
    :param rows:
    :param reverse: 是否倒序
    :return:
    """
    reader = pd.read_csv(csv_file, sep='\t', iterator=True)
    try:
        df = reader.get_chunk(rows)
        if reverse:
            df = df.iloc[::-1]
        df.to_csv(save_csv, index=False)
    except StopIteration:
        print("Iteration is stopped.")
    # reader = pd.read_csv(csv_file, error_bad_lines=False, nrows=100)
    # save_data = reader.iloc[1: 90]
    # save_data.to_csv(save_file, index=False)


def write_row_csv(csv_path, data):
    """
    按行写入csv文件
    :param csv_path: csv文件
    :param data: 二维数组 [[1, 2], [1, 2, 3]]
    :return:
    """
    with open(csv_path, "w") as file:
        writer = csv.writer(file)
        writer.writerows(data)


if __name__ == "__main__":
    # read_excel_merge_cell(file_path=Path(r"C:\Users\zhousf-a\Desktop\images_40_1_ocr_data.xlsx"),
    #                       tmp_excel=Path(r"C:\Users\zhousf-a\Desktop\images_40_1_ocr_data-tmp.xlsx"),
    #                       delete_duplicates_rate=0.85)
    pass



# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function:
import re


def get_digit_char(string: str) -> str:
    """
    提取数字
    '84.2' -> '842'
    'abc123' -> '123'
    :param string:
    :return:
    """
    return re.sub(u"([^\u0030-\u0039])", "", string)


def get_integer_decimals(string: str) -> list:
    """
    提取小数和整数
    'abc84.2'    -> ['84.2']
    'abc84.2@56' -> ['84.2', '56']
    :param string:
    :return:
    """
    res = re.findall(r"\d+[.\d]*", string)
    return res


def only_contain_letter_char(self, string: str) -> bool:
    """
    仅包含字母（大小写）
    """
    return len(self.get_letter_char(string)) == len(string)


def get_letter_char(string: str) -> str:
    # 提取大小写字母
    return re.sub(u"([^\u0041-\u005a\u0061-\u007a])", "", string)


def get_digit_letter_char(string: str) -> str:
    # 提取大小写字母、数字
    return re.sub(u"([^\u0041-\u005a\u0061-\u007a\u0030-\u0039])", "", string)


def only_chinese(string: str) -> bool:
    """
    string都是中文
    :param string:
    :return: True都是中文 | 否
    """
    match_chinese = re.sub(u"([^\u4e00-\u9fa5])", "", string)
    return len(match_chinese) == len(string)


def normalize_cos_sign(string, sign: str = None):
    """
    cos符号标准化：将cos15°转成cos(radians(15))
    cos15° -> cos(radians(15))
    cos5 -> cos(radians(5))
    cosa° -> cos(radians(a))
    cosθ -> cos(radians(θ))
    cos(a) -> cos(radians(a))
    cos(15°) -> cos(radians(15))
    cos(5) -> cos(radians(5))
    :param string:
    :param sign:
    :return:
    """
    if not sign:
        sign = "radians"
    # ori = string
    items = re.split("[+-/*]", string)
    if len(items) > 0:
        for item in items:
            res = re.findall("[Cc][Oo0][Ss]", item)
            if len(res) > 0:
                value = item.replace(res[0], "").replace("°", "").replace("。", "").replace("(", "").replace(")", "")
                if sign:
                    value = "({0}({1}))".format(sign, value)
                else:
                    value = "({0})".format(value)
                string = string.replace(item, "cos" + value)
    # print(ori, "->", side_length)
    return string


def normalize_multiple_sign(string, from_char: list = None, to_char="*"):
    """
    乘号标准化：将x、X、×转成*
    161.9+x2-nXSxd+nxS+2x+2x(a+b) -> 161.9+x2-n*S*d+n*S+2x+2*(a+b)
    :param string:
    :param from_char: ["x", "X"]
    :param to_char: 标准化的字符
    :return:
    """
    if from_char is None:
        from_char = ["x", "X"]
    items = re.split("[+-/*]", string)
    if len(items) > 0:
        for item in items:
            from_join = "".join(from_char)
            pattern = "^[^"+from_join+"].*["+from_join+"]{1}.*[^"+from_join+"]$"
            res = re.findall(pattern, item)
            if len(res) > 0:
                item = str(res[0]).replace("x", to_char).replace("X", to_char).replace("×", to_char)
                string = string.replace(res[0], item)
    return string


def normalize_calculate_formula(string: str) -> str:
    """
    计算公式标准化，字符串排序
    (320+10707/2+1272)*2+(12+11)  ->   2*(10707/2+1272+320)+(11+12)
    320+10707/2+1272              ->   10707/2+1272+320
    320+1272*2                    ->   1272*2+320
    320+1272*2+(12+11)            ->   1272*2+320+(11+12)
    17+3.1416*110                 ->   110*3.1416+17
    """
    # 先把括号里的顺序排列好
    res = re.findall(r"[(](.*?)[)]", string)
    string_blank = string
    blank_map = {}
    if len(res):
        for i, item in enumerate(res):
            ite = str(item).split("+")
            ite.sort()
            string = string.replace(item, "+".join(ite))
            string_blank = string_blank.replace("(" + item + ")", "A{0}".format(i))
            blank_map["A{0}".format(i)] = "(" + "+".join(ite) + ")"
    else:
        ite = string_blank.split("+")
        ite.sort()
        string_blank = "+".join(ite)
    # 再把乘号顺序排列好
    res = string_blank.split("+")
    res.sort()
    string_blank = "+".join(res)
    for i, item in enumerate(res):
        ite = item.split("*")
        ite.sort()
        string_blank = string_blank.replace(item, "*".join(ite))
    # 再次排序
    res = string_blank.split("+")
    res.sort()
    string_blank = "+".join(res)
    # 最后映射转换
    for blank in blank_map:
        string_blank = string_blank.replace(blank, blank_map.get(blank))
    return string_blank


# -*- coding:utf-8 -*-
# Author:      zhousf
# Description: 半角、全角转换

def half_to_full_char(uchar):
    """
    单个字符 半角转全角
    :param uchar:
    :return:
    """
    inside_code = ord(uchar)
    # 不是半角字符就返回原来的字符
    if inside_code < 0x0020 or inside_code > 0x7e:
        return uchar
    # 除了空格其他的全角半角的公式为: 半角 = 全角 - 0xfee0
    if inside_code == 0x0020:
        inside_code = 0x3000
    else:
        inside_code += 0xfee0
    return chr(inside_code)


def full_to_half_char(uchar):
    """
    单个字符 全角转半角
    :param uchar:
    :return:
    """
    inside_code = ord(uchar)
    if inside_code == 0x3000:
        inside_code = 0x0020
    else:
        inside_code -= 0xfee0
    # 转完之后不是半角字符返回原来的字符
    if inside_code < 0x0020 or inside_code > 0x7e:
        return uchar
    return chr(inside_code)


def is_half_number(uchar):
    """
    判断一个unicode是否是半角数字
    :param uchar:
    :return:
    """
    if u'\u0030' <= uchar <= u'\u0039':
        return True
    else:
        return False


def is_full_number(uchar):
    """
    判断一个unicode是否是全角数字
    :param uchar:
    :return:
    """
    if u'\uff10' <= uchar <= u'\uff19':
        return True
    else:
        return False


def is_half_alphabet(uchar):
    """
    判断一个unicode是否是半角英文字母
    :param uchar:
    :return:
    """
    if (u'\u0041' <= uchar <= u'\u005a') or (u'\u0061' <= uchar <= u'\u007a'):
        return True
    else:
        return False


def is_full_alphabet(uchar):
    """
    判断一个unicode是否是全角英文字母
    :param uchar:
    :return:
    """
    if (u'\uff21' <= uchar <= u'\uff3a') or (u'\uff41' <= uchar <= u'\uff5a'):
        return True
    else:
        return False


def full_to_half_str(ustring):
    """
    字符串全角转半角
    :param ustring:
    :return:
    """
    return "".join([full_to_half_char(uchar) for uchar in ustring])
